"""Custom exceptions."""


def get_adobe_exception(exception):
    try:
        status_code = exception.response.status_code
        json_data = exception.response.json()
        reason = json_data['code']
        content = json_data['message']
        if AdobeSignNoMoreSignerException.is_no_more_signer(status_code,
                                                            reason):
            return AdobeSignNoMoreSignerException(content, reason,
                                                  cause=exception)

        if AdobeSignInvalidAccessTokenException.is_invalid_token(
                status_code, reason):
            return AdobeSignInvalidAccessTokenException(content,
                                                        cause=exception)
        if AdobeSignInvalidUserException.is_invalid_user(status_code, reason):
            return AdobeSignInvalidUserException(content, cause=exception)

        message = '{} {} {}'.format(exception, reason, content)
    except Exception:
        try:
            message = '{} {} {}'.format(exception,
                                        status_code,
                                        exception.response.body)
        except Exception:
            message = exception
    return AdobeSignException(message, cause=exception)


class AdobeSignException(Exception):
    """An error occurred with AdobeSign wrapper API."""

    def __init__(self, message, cause=None):
        super(AdobeSignException, self).__init__(message)
        self.__cause__ = cause


class AdobeSignNoMoreSignerException(AdobeSignException):
    CODE_REASON = ('AGREEMENT_EXPIRED', 'AGREEMENT_NOT_SIGNABLE',
                   'AGREEMENT_NOT_VISIBLE')

    def __init__(self, message, reason, cause=None):
        if reason not in AdobeSignNoMoreSignerException.CODE_REASON:
            reason = 'Unexpected reason: {}'.format(reason)
        self.reason = reason
        super(AdobeSignNoMoreSignerException, self).__init__(message, cause)

    @staticmethod
    def is_no_more_signer(status_code, reason):
        return status_code == 404 and \
               reason in AdobeSignNoMoreSignerException.CODE_REASON


class AdobeSignInvalidAccessTokenException(AdobeSignException):
    def __init__(self, message, cause=None):
        super(AdobeSignInvalidAccessTokenException, self).__init__(message,
                                                                   cause)

    @staticmethod
    def is_invalid_token(status_code, reason):
        return status_code == 401 and 'INVALID_ACCESS_TOKEN' in reason


class AdobeSignInvalidUserException(AdobeSignException):
    def __init__(self, message, cause=None):
        super(AdobeSignInvalidUserException, self).__init__(message, cause)

    @staticmethod
    def is_invalid_user(status_code, reason):
        return status_code == 401 and 'INVALID_USER' in reason



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Python packaging."""
import os

from setuptools import setup, find_packages

#: Absolute path to directory containing setup.py file.
here = os.path.abspath(os.path.dirname(__file__))

README = open(os.path.join(here, 'README.rst')).read()
VERSION = open(os.path.join(here, 'VERSION')).read().strip()

REQUIREMENTS = [
    'Django>=1.11,<2.1',
    'django-anysign>=1.2',
    'requests',
    'requests_oauthlib'
]

if __name__ == '__main__':  # Do not run setup() when we import this module.
    setup(
        name='django-adobesign',
        version=VERSION,
        description='Django application to manage online signature with '
                    'AdobeSign.',
        long_description=README,
        classifiers=[
            "Operating System :: POSIX",
            'Framework :: Django',
            'Framework :: Django :: 1.11',
            'Framework :: Django :: 2.0',
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
        ],
        keywords='peopledoc anysign adobesign',
        author='Peopledoc',
        author_email='david.steinberger@people-doc.com',
        url='http://github.com/peopledoc/django-adobesign',
        packages=find_packages(exclude=['demo']),
        include_package_data=True,
        zip_safe=True,
        install_requires=REQUIREMENTS,
    )



from __future__ import unicode_literals

from django.db import transaction
from django.views.generic.base import RedirectView
from django.views.generic.detail import SingleObjectMixin
from django_anysign import api as django_anysign

from django_adobesign.exceptions import AdobeSignException


class SignerReturnView(SingleObjectMixin, RedirectView):
    """Handle return of signer on project after document signing/reject.
    """
    permanent = False

    def get_redirect_url(self, *args, **kwargs):
        """Route request to signer return view depending on status.
        Trigger events for latest signer: calls
        ``signer_{status}`` methods.

        Adobe signer status:
        - COMPLETED: everybody has signed
        - WAITING_FOR_OTHERS: the signer has already signed
        - NOT_YET_VISIBLE: the signer signing url is not generated yet,
                           it is not his turn to sign
        - WAITING_FOR_MY_SIGNATURE: the user signing url is generated, it is
                                    his turn to sign
        - CANCELLED: the user has refused to sign
        -  WAITING_FOR_AUTHORING, WAITING_FOR_MY_DELEGATION
            WAITING_FOR_MY_ACCEPTANCE, WAITING_FOR_MY_ACKNOWLEDGEMENT
            WAITING_FOR_MY_APPROVAL, WAITING_FOR_MY_FORM_FILLING
        """

        signer = self.get_current_signer()
        if not signer:
            return self.get_signer_error_url('No more signer')
        agreement_id = self.signature.signature_backend_id
        adobe_signer_id = signer.signature_backend_id
        status = self.backend.get_signer_status(agreement_id, adobe_signer_id)
        if status == 'WAITING_FOR_MY_SIGNATURE':
            return self.get_signer_error_url('User mismatch')
        if status == 'CANCELLED':
            self.signer_cancelled(signer, status)
            return self.get_signer_canceled_url(status)

        if status == 'COMPLETED':
            if not self.backend.is_last_signer(signer):
                raise AdobeSignException('Consistency issue, agreement {} '
                                         'is complete, remaining signers have '
                                         'been found'.format(agreement_id))
            self.signer_signed(status, signer)
            self.update_signature(status)
            return self.get_signer_signed_url(status)
        if status == 'WAITING_FOR_OTHERS':
            self.signer_signed(status, signer)
            return self.get_signer_signed_url(status)

        self.update_signer(signer, status)
        return self.get_signer_error_url()

    def get_current_signer(self):
        for signer in self.signature.signers.all().order_by('signing_order'):
            if not self.has_already_signed(signer):
                return signer

    def get_queryset(self):
        model = django_anysign.get_signature_model()
        return model.objects.all()

    def get_signed_document(self):
        # In our model, there is only one doc.
        return next(
            self.backend.get_documents(self.signature.signature_backend_id))

    def signer_cancelled(self, signer,  status):
        """Handle 'Cancel' status for signer."""
        self.update_signer(signer=signer, status=status)
        self.update_signature(status=status)

    def signer_signed(self, status, signer):
        """ Update signer status after he sign
        """
        # download signed document out of the atomic block
        signed_document = self.get_signed_document()
        with transaction.atomic():
            self.replace_document(signed_document)
            self.update_signer(signer, status)

    @property
    def signature(self):
        """Signature model instance.

        This is a shortcut property using a cache.
        If you want to adapt the implementation, consider overriding
        :meth:`get_signature`.

        """
        try:
            return self._signature
        except AttributeError:
            self._signature = self.get_object()
            return self._signature

    @property
    def backend(self):
        try:
            return self._backend
        except AttributeError:
            self._backend = self.signature.signature_backend
            return self._backend

    def has_already_signed(self, signer):
        raise NotImplementedError()

    def get_signer_canceled_url(self, status):
        """Url redirect when signer canceled signature."""
        raise NotImplementedError()

    def get_signer_error_url(self, message=''):
        """Url redirect when failure."""
        raise NotImplementedError()

    def get_signer_signed_url(self, status):
        """Url redirect when signer signed signature."""
        raise NotImplementedError()

    def update_signature(self, status):
        """ Update signature with ``status``."""
        raise NotImplementedError()

    def update_signer(self, signer, status, message=''):
        """Update ``signer`` with ``status``."""
        raise NotImplementedError()

    def replace_document(self, signed_document):
        """Replace original document by signed one."""
        raise NotImplementedError()



from django_anysign import api as django_anysign


class AdobeSignBackend(django_anysign.SignatureBackend):
    def __init__(self, adobesign_client, name='AdobeSign', code='adobesign',
                 url_namespace='anysign'):
        """Setup.

        Additional keyword arguments are passed to
        :class:`~django_adobesign.client.AdobeSignCient` constructor, in order
        to setup :attr:`adobesign client`.

        """
        super(AdobeSignBackend, self).__init__(
            name=name,
            code=code,
            url_namespace=url_namespace,
        )
        self.adobesign_client = adobesign_client

    def get_adobesign_participants(self, signature):
        """Return list of AdobeSign's Signers for Signature instance.

        Default implementation reads name and email from database.

        """
        jsonified_signers = []
        signers = signature.signers.all().order_by('signing_order')
        for position, signer in enumerate(signers, start=1):
            signer = self.adobesign_client.jsonify_participant(
                name=signer.full_name,
                email=signer.email,
                order=position
            )
            jsonified_signers.append(signer)
        return jsonified_signers

    def create_signature(self, signature, post_sign_redirect_url=None,
                         post_sign_redirect_delay=0,
                         send_mail=True, **extra_data):
        """Register ``signature`` in AdobeSign service, return updated object.
        This method calls ``save()`` on ``signature`` and ``signer``.

        """
        document = next(signature.signature_documents())

        # Upload document
        response = self.adobesign_client.upload_document(document)
        transient_document_id = response.get('transientDocumentId')
        result = self.adobesign_client.post_agreement(
            transient_document_id=transient_document_id,
            name=str(signature),
            participants=self.get_adobesign_participants(signature),
            post_sign_redirect_url=post_sign_redirect_url,
            post_sign_redirect_delay=post_sign_redirect_delay,
            send_mail=send_mail,
            **extra_data)

        # Update signature instance with record_id
        signature.signature_backend_id = result['id']
        signature.save(update_fields=['signature_backend_id'])

        # Update signers instance with external id
        self.map_adobe_signer_to_signer(signature)
        return signature

    def update_signer_status(self, signer, status):
        raise NotImplementedError()

    def map_adobe_signer_to_signer(self, signature):
        # Can raise a Signer.DoesNotExist
        for adobe_signer in self.get_all_signers(
                signature.signature_backend_id).get('participantSets', []):
            # We only have 1 signer by turn
            email = adobe_signer['memberInfos'][0]['email'].lower()
            signer = signature.signers.get(signing_order=adobe_signer['order'],
                                           email=email)
            signer.signature_backend_id = adobe_signer['id']
            signer.save(update_fields=['signature_backend_id'])
            self.update_signer_status(signer, adobe_signer['status'])

    def get_agreements(self, page_size=20, cursor=None, **extra_params):
        """
            Return all agreements associated to the given access token

            Be careful: reply structure is not the same if there is agreements
            or not

            case not empty: {
                            "page": {"nextCursor": "..."}
                            "userAgreementList": [ agreement1, ... ]
                            }
            case empty: []

            For compliance, in case of empty agreement list we artificially
            return the structure above
        """
        agreements = self.adobesign_client.get_agreements(page_size, cursor,
                                                          **extra_params)
        return agreements or {'userAgreementList': [], 'page': {}}

    def get_next_signers(self, agreement_id):
        """ Return the next signer list."""
        members = self.adobesign_client. \
            get_members(agreement_id, include_next_participant_set=True)
        return members.get('nextParticipantSets', [])

    def get_next_signer_urls(self, agreement_id):
        """
            Return an array of urls for current signer set
        """
        return self.adobesign_client.get_signing_url(agreement_id)

    def get_next_signer_url(self, agreement_id):
        """
            Return the first next signer url and mail if exists
        """
        next_signers_url = self.get_next_signer_urls(agreement_id)
        if 'signingUrlSetInfos' in next_signers_url:
            set_infos = next_signers_url['signingUrlSetInfos']
            if set_infos and set_infos[0]['signingUrls']:
                return set_infos[0]['signingUrls'][0]['email'], \
                       set_infos[0]['signingUrls'][0]['esignUrl']
        return None, None

    def get_all_signers(self, agreement_id):
        """
            Return the list of all signers info
        """
        return self.adobesign_client. \
            get_members(agreement_id, include_next_participant_set=False)

    def get_signer(self, argeement_id, signer_id):
        return self.adobesign_client.get_signer(argeement_id, signer_id)

    def get_signer_status(self, argeement_id, signer_id):
        signer = self.get_signer(argeement_id, signer_id)
        return signer.get('status')

    def is_last_signer(self, signer):
        """Return True if ``signer`` is the last signer for the signature
        request.
        """
        return not signer.signature.signers.filter(
            signing_order__gt=signer.signing_order).exists()

    def get_documents(self, agreement_id):
        documents_info = self.adobesign_client.get_documents(agreement_id)
        for doc_info in documents_info.get('documents', []):
            yield self.adobesign_client.get_document(agreement_id,
                                                     doc_info['id'])

    def get_refuse_comment(self, agreement_id):
        """
        Return the refuse comment from agreement
        """
        events = self.adobesign_client.get_events(agreement_id)
        for event in events.get('events'):
            if event.get('type') == "REJECTED":
                # comment is mandatory
                return event["comment"]



"""Declaration of API shortcuts.

Everything declared (or imported) in this module is exposed in
:mod:`django_opentrust.api` root package, i.e. available when one does
``from django_opentrust import api as django_opentrust``.

Here are the motivations of such an "api" module:

* as a `django-opentrust` library user, in order to use `django-opentrust`, I
  just do ``from django_opentrust import api as django_opentrust``.
  It is enough for most use cases. I do not need to bother with more
  `django_opentrust` internals. I know this API will be maintained,
  documented, and not deprecated/refactored without notice.

* as a `django-opentrust` library developer, in order to maintain
  `django-opentrust` API, I focus on things declared in
  :mod:`django_opentrust.api`. It is enough. It is required. I take care of
  this API. If there is a change in this API between consecutive releases, then
  I use :class:`DeprecationWarning` and I mention it in release notes.

It also means that things not exposed in :mod:`django_opentrust.api` are not
part of the deprecation policy. They can be moved, changed, removed without
notice.

"""
from django_adobesign.backend import AdobeSignBackend
from django_adobesign.client import AdobeSignClient
from django_adobesign.client import AdobeSignOAuthSession
from django_adobesign.exceptions import AdobeSignException


__all__ = [
    'AdobeSignBackend',
    'AdobeSignClient',
    'AdobeSignException',
    'AdobeSignOAuthSession',
]
